from __future__ import annotations

import json

import requests
from soda.execution.partition import Partition
from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef
from soda.sampler.sampler import Sampler


class HTTPSampler(Sampler):
    def __init__(self, url: str, format: str = "json", link: str | None = None, message: str = "", link_text: str = ""):
        self.url = url
        self.format = format
        self.link = link
        self.message = message
        self.link_text = link_text

    def store_sample(self, sample_context: SampleContext) -> SampleRef | None:
        self.logs.info(f"Sending failed row samples to {self.url}")
        sample_rows = sample_context.sample.get_rows()
        row_count = sample_context.sample.get_rows_count()
        sample_schema = sample_context.sample.get_schema()
        link = self.link
        message = f"{self.message}"
        link_text = self.link_text

        if row_count != 0:
            result_dict = {
                "schema": sample_schema.get_dict(),
                "count": row_count,
                "rows": sample_rows,
                "datasource": sample_context.sample.data_source.data_source_name,
                "dataset": Partition.get_table_name(sample_context.partition),
                "scan_definition": sample_context.scan._scan_definition_name,
                "check_name": sample_context.check_name,
            }

            response = requests.post(self.url, json=json.dumps(result_dict, sort_keys=True, default=str))
            message = f"{self.message} {response.text}"

            if response.status_code != 200:
                self.logs.error(f"Unable to upload failed rows to {self.url} -  {response.text}")
            else:
                self.logs.info(f"Uploaded {row_count} failed rows to {self.url}")
        else:
            link = None
            message = None
            link_text = None
            self.logs.info(f"No failed rows recognized. Nothing uploaded to {self.url}")

        if sample_context.samples_limit is not None:
            stored_row_count = row_count if row_count < sample_context.samples_limit else sample_context.samples_limit
        else:
            stored_row_count = row_count

        return SampleRef(
            name=sample_context.sample_name,
            schema=sample_schema,
            total_row_count=row_count,
            stored_row_count=stored_row_count,
            type=SampleRef.TYPE_NOT_PERSISTED,
            link=link,
            message=message,
            link_text=link_text,
        )



from __future__ import annotations

from soda.execution.metric.query_metric import QueryMetric
from soda.sampler.sample_ref import SampleRef


class ReferenceMetric(QueryMetric):
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        check: ReferenceCheck,
        partition: Partition,
        single_source_column: Column,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=single_source_column,
            name="reference",
            check=check,
            identity_parts=[
                check.check_cfg.source_column_names,
                check.check_cfg.target_table_name,
                check.check_cfg.target_column_names,
            ],
        )
        self.check = check
        self.invalid_references_sample_ref: Optional[SampleRef] = None

    def __str__(self):
        return f'"{self.name}"'

    def ensure_query(self):
        from soda.execution.query.reference_query import ReferenceQuery

        reference_query = ReferenceQuery(
            data_source_scan=self.data_source_scan,
            metric=self,
            samples_limit=self.samples_limit,
            partition=self.partition,
        )
        self.data_source_scan.queries.append(reference_query)
        self.queries.append(reference_query)



from __future__ import annotations

from typing import TYPE_CHECKING

from soda.execution.query.query import Query
from soda.profiling.profile_columns_result import ProfileColumnsResultColumn

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan
    from soda.sodacl.data_source_check_cfg import ProfileColumnsCfg


class TextColumnProfiler:
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        profile_columns_cfg: ProfileColumnsCfg,
        table_name: str,
        column_name: str,
        column_data_type: str,
    ) -> None:
        self.data_source_scan = data_source_scan
        self.data_source = data_source_scan.data_source
        self.logs = data_source_scan.scan._logs
        self.profile_columns_cfg = profile_columns_cfg
        self.table_name = table_name
        self.column_name = column_name
        self.column_data_type = column_data_type
        self.result_column = ProfileColumnsResultColumn(column_name=column_name, column_data_type=column_data_type)

    def profile(self) -> ProfileColumnsResultColumn:
        self.logs.debug(f"Profiling column {self.column_name} of {self.table_name}")

        # frequent values for text column
        self._set_result_column_value_frequency_attribute()

        # pure text aggregates
        self._set_result_column_text_aggregation_attributes()

        return self.result_column

    def _set_result_column_value_frequency_attribute(self) -> None:
        value_frequencies = self._compute_value_frequency()
        if value_frequencies:
            self.result_column.set_frequency_metric(value_frequencies)
        else:
            self.logs.warning(
                f"Database returned no results for textual frequent values in {self.table_name}, column: {self.column_name}"
            )

    def _set_result_column_text_aggregation_attributes(self) -> None:
        text_aggregates = self._compute_text_aggregates()
        if text_aggregates:
            self.result_column.set_text_aggregation_metrics(text_aggregates)
        else:
            self.logs.warning(
                f"Database returned no results for textual aggregates in {self.table_name}, column: {self.column_name}"
            )

    def _compute_value_frequency(self) -> list[tuple] | None:
        # frequent values for text column
        value_frequencies_sql = self.data_source.profiling_sql_values_frequencies_query(
            "text",
            self.table_name,
            self.column_name,
            self.profile_columns_cfg.limit_mins_maxs,
            self.profile_columns_cfg.limit_frequent_values,
        )
        value_frequencies_query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"profiling-{self.table_name}-{self.column_name}-value-frequencies-text",
            sql=value_frequencies_sql,
        )
        value_frequencies_query.execute()
        frequency_rows = value_frequencies_query.rows
        return frequency_rows

    def _compute_text_aggregates(self) -> list[tuple] | None:
        text_aggregates_sql = self.data_source.profiling_sql_aggregates_text(self.table_name, self.column_name)
        text_aggregates_query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"profiling: {self.table_name}, {self.column_name}: get textual aggregates",
            sql=text_aggregates_sql,
        )
        text_aggregates_query.execute()
        text_aggregates_rows = text_aggregates_query.rows
        return text_aggregates_rows



from typing import Dict

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.metric import Metric
from soda.execution.metric.numeric_query_metric import NumericQueryMetric
from soda.execution.partition import Partition
from soda.execution.query.user_defined_failed_rows_expression_query import (
    UserDefinedFailedRowsExpressionQuery,
)

KEY_FAILED_ROWS_COUNT = "failed_rows_count"


class UserDefinedFailedRowsExpressionCheck(Check):
    def __init__(
        self,
        check_cfg: "UserDefinedFailedRowsExpressionCheckCfg",
        data_source_scan: "DataSourceScan",
        partition: Partition,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
        )
        self.check_value = None
        self.metrics[KEY_FAILED_ROWS_COUNT] = self.data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=self.data_source_scan,
                partition=partition,
                column=None,
                metric_name="row_count",
                metric_args=None,
                filter=check_cfg.fail_condition_sql_expr,
                aggregation=None,
                check_missing_and_valid_cfg=None,
                column_configurations_cfg=None,
                check=self,
            )
        )

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        self.check_value: int = metrics.get(KEY_FAILED_ROWS_COUNT).value

        # Thresholds path
        if self.check_cfg.fail_threshold_cfg or self.check_cfg.warn_threshold_cfg:
            if self.check_cfg.fail_threshold_cfg and self.check_cfg.fail_threshold_cfg.is_bad(self.check_value):
                self.outcome = CheckOutcome.FAIL
            elif self.check_cfg.warn_threshold_cfg and self.check_cfg.warn_threshold_cfg.is_bad(self.check_value):
                self.outcome = CheckOutcome.WARN
            else:
                self.outcome = CheckOutcome.PASS
        else:
            # Original non-threshold path
            if self.check_value > 0:
                self.outcome = CheckOutcome.FAIL
            else:
                self.outcome = CheckOutcome.PASS

        if self.check_value > 0:
            # Collect failed rows
            failed_rows_sql = self.get_failed_rows_sql()
            failed_rows_query = UserDefinedFailedRowsExpressionQuery(
                data_source_scan=self.data_source_scan,
                check_name=self.check_cfg.source_line,
                sql=failed_rows_sql,
                samples_limit=self.check_cfg.samples_limit,
                partition=self.partition,
                metric=self.metrics[KEY_FAILED_ROWS_COUNT],
            )
            failed_rows_query.execute()
            if failed_rows_query.sample_ref:
                self.failed_rows_sample_ref = failed_rows_query.sample_ref

    def get_failed_rows_sql(self) -> str:
        columns = ", ".join(
            self.data_source_scan.data_source.sql_select_all_column_names(self.partition.table.table_name)
        )
        sql = (
            f"SELECT {columns} \n"
            f"FROM {self.partition.table.qualified_table_name} \n"
            f"WHERE ({self.check_cfg.fail_condition_sql_expr})"
        )
        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            condition = scan.jinja_resolve(definition=partition_filter, location=self.check_cfg.location)
            sql += f"\n      AND ({condition})"
        return sql

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = {"check_value": self.check_value}
        if self.failed_rows_sample_ref:
            log_diagnostics["failed_rows_sample_ref"] = str(self.failed_rows_sample_ref)
        return log_diagnostics



from __future__ import annotations


class Location:
    def __init__(
        self,
        file_path: str | None = None,
        line: int | None = None,
        col: int | None = None,
    ):
        self.file_path = file_path
        self.line = line
        self.col = col

    def __str__(self):
        return f"line={self.line},col={self.col} in {self.file_path}"

    def __hash__(self) -> int:
        return hash((self.file_path, self.line, self.col))

    def get_identity_parts(self) -> list:
        return [self.file_path, self.line]

    def get_cloud_dict(self):
        return {
            "filePath": self.file_path,
            "line": self.line,
            "col": self.col,
        }

    def get_dict(self):
        return {
            "filePath": self.file_path,
            "line": self.line,
            "col": self.col,
        }



from __future__ import annotations

from soda.scan import Scan


class DataSourceUtils:
    """
    EXPERIMENTAL: Utility API for DataSource and connections.
    """

    def __init__(self, config_path: str | None = None, config_string: str | None = None):
        """
        All the features for config files (e.g. env vars) are supported.
        :param config_path: The directory path containing the data source config yaml files.
        :param config_string: The data source config yaml string.
        """
        self._scan = Scan()
        self.config_path = config_path
        self.config_string = config_string

        if config_path:
            self._scan.add_configuration_yaml_files(self.config_path)

        if config_string:
            self._scan.add_configuration_yaml_str(self.config_string)

    def get_connection(self, data_source_name: str):
        """
        Returns a DB API connection object configured for the given data source name
        :param data_source_name: Name of a data source configured in one of the self.config_path files.
        :return: connection object
        """
        ds = self._scan._data_source_manager.get_data_source(data_source_name)
        if ds is None:
            raise Exception(self._scan.get_logs_text())
        else:
            ds.connect()
            return ds.connection



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from importlib.util import find_spec
from numbers import Number

from soda.execution.metric.query_metric import QueryMetric
from soda.execution.query.sample_query import SampleQuery
from soda.sodacl.format_cfg import FormatHelper


class NumericQueryMetric(QueryMetric):
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        partition: Partition | None,
        column: Column | None,
        metric_name: str,
        metric_args: list[object] | None,
        filter: str | None,
        aggregation: str | None,
        check_missing_and_valid_cfg: MissingAndValidCfg,
        column_configurations_cfg: ColumnConfigurationsCfg,
        check: Check,
    ):
        from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg

        merged_missing_and_valid_cfg = MissingAndValidCfg.merge(check_missing_and_valid_cfg, column_configurations_cfg)
        other_metric_args = metric_args[1:] if isinstance(metric_args, list) and len(metric_args) > 1 else None
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name=metric_name,
            check=check,
            identity_parts=[
                other_metric_args,
                filter,
                merged_missing_and_valid_cfg.get_identity_parts() if merged_missing_and_valid_cfg else None,
            ],
        )

        self.metric_args: list[object] | None = metric_args
        self.filter: str | None = filter
        self.aggregation: str | None = aggregation
        self.missing_and_valid_cfg: MissingAndValidCfg = merged_missing_and_valid_cfg

        # Implementation and other non-identity fields
        self.logs = data_source_scan.scan._logs
        self.column_name = column.column_name if column else None
        self.check = check

    def get_sql_aggregation_expression(self) -> str | None:
        data_source = self.data_source_scan.data_source

        """
        Returns an aggregation SQL expression for the given metric as a str or None if It is not an aggregation metric
        """
        if self.name in ["row_count", "missing_count", "valid_count", "invalid_count"]:
            # These are the conditional count metrics
            condition = None

            if "missing_count" == self.name:
                condition = self.build_missing_condition()

            if "valid_count" == self.name:
                condition = self.build_non_missing_and_valid_condition()

            if "invalid_count" == self.name:
                missing_condition = self.build_missing_condition()
                valid_condition = self.build_valid_condition()
                invalid_condition = self.build_invalid_condition()
                if valid_condition:
                    condition = f"NOT ({missing_condition}) AND NOT ({valid_condition})"
                elif invalid_condition:
                    condition = f"NOT ({missing_condition}) AND ({invalid_condition})"
                else:
                    self.logs.warning(
                        f'Counting invalid without valid or invalid specification does not make sense. ("{self.check.check_cfg.source_line}" @ {self.check.check_cfg.location})'
                    )
                    condition = self.data_source_scan.data_source.expr_false_condition()

            if self.filter:
                condition = f"({self.filter}) AND ({condition})" if condition else self.filter

            if condition:
                return data_source.expr_count_conditional(condition=condition)
            else:
                return data_source.expr_count_all()

        if self.aggregation:
            if self.is_missing_or_validity_configured():
                self.logs.info(
                    f"Missing and validity in {self.column.column_name} will not be applied to custom aggregation expression ({self.aggregation})"
                )
            if self.filter:
                self.logs.error(
                    f"Filter ({self.filter}) can't be applied in combination with a custom metric aggregation expression ({self.aggregation})"
                )
            return self.aggregation

        values_expression = self.column.column_name

        condition_clauses = []
        if self.is_missing_or_validity_configured():
            condition_clauses.append(self.build_non_missing_and_valid_condition())
        if self.filter:
            condition_clauses.append(self.filter)
        if condition_clauses:
            condition = " AND ".join(condition_clauses)
            values_expression = data_source.expr_conditional(condition=condition, expr=self.column_name)

        numeric_format = self.get_numeric_format()
        if numeric_format:
            values_expression = data_source.cast_text_to_number(values_expression, numeric_format)

        return data_source.get_metric_sql_aggregation_expression(self.name, self.metric_args, values_expression)

    def set_value(self, value):
        if value is None or isinstance(value, int):
            self.value = value
        elif self.name in [
            "row_count",
            "missing_count",
            "invalid_count",
            "valid_count",
            "duplicate_count",
        ]:
            self.value = int(value)
        elif isinstance(value, Decimal):
            self.value = float(value)
        # Check if numpy is installed
        elif find_spec("pandas") and find_spec("numpy"):
            # If pandas then numpy is installed already
            import numpy as np
            import pandas as pd

            if isinstance(value, np.datetime64):
                # Convert numpy datetime64 to python datetime
                self.value = datetime.utcfromtimestamp(value.tolist() / 1e9)
            elif isinstance(value, pd.Timestamp):
                self.value = value.to_pydatetime()
            elif isinstance(value, np.floating):
                self.value = float(value)
            elif isinstance(value, np.integer):
                self.value = int(value)
            else:
                self.value = value
        else:
            self.value = value

    def ensure_query(self):
        self.partition.ensure_query_for_metric(self)

    def build_missing_condition(self) -> str:
        from soda.execution.data_source import DataSource

        column_name = self.column_name
        data_source: DataSource = self.data_source_scan.data_source

        def append_missing(missing_and_valid_cfg):
            if missing_and_valid_cfg:
                if missing_and_valid_cfg.missing_values:
                    sql_literal_missing_values = data_source.literal_list(missing_and_valid_cfg.missing_values)
                    validity_clauses.append(f"{column_name} IN {sql_literal_missing_values}")

                missing_format = missing_and_valid_cfg.missing_format
                if missing_format:
                    missing_expression = self.data_source_scan.data_source.get_default_format_expression(
                        column_name, missing_format, missing_and_valid_cfg.missing_format_location
                    )
                    if missing_expression:
                        validity_clauses.append(missing_expression)

                if missing_and_valid_cfg.missing_regex:
                    missing_regex = data_source.escape_regex(missing_and_valid_cfg.missing_regex)
                    validity_clauses.append(data_source.expr_regexp_like(column_name, missing_regex))

        validity_clauses = [f"{column_name} IS NULL"]

        append_missing(self.missing_and_valid_cfg)

        return " OR ".join(validity_clauses)

    def build_valid_condition(self) -> str | None:
        column_name = self.column_name
        data_source = self.data_source_scan.data_source

        def append_valid(missing_and_valid_cfg):
            if missing_and_valid_cfg:
                if missing_and_valid_cfg.valid_values:
                    valid_values_sql = data_source.literal_list(missing_and_valid_cfg.valid_values)
                    in_expr = data_source.expr_in(column_name, valid_values_sql)
                    validity_clauses.append(in_expr)

                valid_format = missing_and_valid_cfg.valid_format
                if valid_format:
                    validity_expression = self.data_source_scan.data_source.get_default_format_expression(
                        column_name, valid_format, missing_and_valid_cfg.valid_format_location
                    )
                    if validity_expression:
                        validity_clauses.append(validity_expression)
                if missing_and_valid_cfg.valid_regex is not None:
                    valid_regex = data_source.escape_regex(missing_and_valid_cfg.valid_regex)
                    regex_like_expr = data_source.expr_regexp_like(column_name, valid_regex)
                    validity_clauses.append(regex_like_expr)
                if missing_and_valid_cfg.valid_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    gte_expr = f"{length_expr} = {missing_and_valid_cfg.valid_length}"
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_min_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    gte_expr = f"{length_expr} >= {missing_and_valid_cfg.valid_min_length}"
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_max_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    lte_expr = f"{length_expr} <= {missing_and_valid_cfg.valid_max_length}"
                    validity_clauses.append(lte_expr)
                if missing_and_valid_cfg.valid_min is not None:
                    gte_expr = get_boundary_expr(missing_and_valid_cfg.valid_min, ">=")
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_max is not None:
                    lte_expr = get_boundary_expr(missing_and_valid_cfg.valid_max, "<=")
                    validity_clauses.append(lte_expr)

        def get_boundary_expr(value, comparator: str):
            valid_format = self.missing_and_valid_cfg.valid_format if self.missing_and_valid_cfg else None
            if valid_format:
                cast_expr = data_source.cast_text_to_number(column_name, valid_format)
                return f"{cast_expr} {comparator} {value}"
            else:
                return f"{column_name} {comparator} {value}"

        validity_clauses = []

        append_valid(self.missing_and_valid_cfg)

        return " AND ".join(validity_clauses) if len(validity_clauses) != 0 else None

    def build_invalid_condition(self) -> str | None:
        column_name = self.column_name
        data_source = self.data_source_scan.data_source

        def append_invalid(missing_and_valid_cfg):
            if missing_and_valid_cfg:
                invalid_format = missing_and_valid_cfg.invalid_format
                if invalid_format:
                    invalidity_expression = self.data_source_scan.data_source.get_default_format_expression(
                        column_name, invalid_format, missing_and_valid_cfg.invalid_format_location
                    )
                    if invalidity_expression:
                        invalidity_clauses.append(invalidity_expression)

                if missing_and_valid_cfg.invalid_values:
                    invalid_values_sql = data_source.literal_list(missing_and_valid_cfg.invalid_values)
                    in_expr = data_source.expr_in(column_name, invalid_values_sql)
                    invalidity_clauses.append(in_expr)

                if missing_and_valid_cfg.invalid_regex is not None:
                    invalid_regex = data_source.escape_regex(missing_and_valid_cfg.invalid_regex)
                    regex_like_expr = data_source.expr_regexp_like(column_name, invalid_regex)
                    invalidity_clauses.append(regex_like_expr)

        invalidity_clauses = []

        append_invalid(self.missing_and_valid_cfg)

        return " OR ".join(invalidity_clauses) if len(invalidity_clauses) != 0 else None

    def build_non_missing_and_valid_condition(self):
        missing_condition = self.build_missing_condition()
        valid_condition = self.build_valid_condition()
        invalid_condition = self.build_invalid_condition()
        if valid_condition:
            return f"NOT ({missing_condition}) AND ({valid_condition})"
        elif invalid_condition:
            return f"NOT ({missing_condition}) AND NOT ({invalid_condition})"
        else:
            return f"NOT ({missing_condition})"

    def get_numeric_format(self) -> str | None:
        if self.missing_and_valid_cfg and FormatHelper.is_numeric(self.missing_and_valid_cfg.valid_format):
            return self.missing_and_valid_cfg.valid_format
        return None

    def is_missing_or_validity_configured(self) -> bool:
        return self.missing_and_valid_cfg is not None

    metric_names_with_failed_rows = ["missing_count", "invalid_count"]

    def create_failed_rows_sample_query(self) -> SampleQuery | None:
        sampler = self.data_source_scan.scan._configuration.sampler
        if (
            sampler
            and self.name in self.metric_names_with_failed_rows
            and isinstance(self.value, Number)
            and self.value > 0
        ):
            where_clauses = []
            passing_where_clauses = []
            partition_filter = self.partition.sql_partition_filter
            if partition_filter:
                resolved_filter = self.data_source_scan.scan.jinja_resolve(definition=partition_filter)
                where_clauses.append(resolved_filter)
                passing_where_clauses.append(resolved_filter)

            if self.name == "missing_count":
                where_clauses.append(f"({self.build_missing_condition()})")
                passing_where_clauses.append(f"NOT ({self.build_missing_condition()})")
            elif self.name == "invalid_count":
                where_clauses.append(f"NOT ({self.build_missing_condition()})")
                passing_where_clauses.append(f"NOT ({self.build_missing_condition()})")

                valid_condition = self.build_valid_condition()
                if valid_condition:
                    where_clauses.append(f"NOT ({valid_condition})")
                    passing_where_clauses.append(f"({valid_condition})")

                invalid_condition = self.build_invalid_condition()
                if invalid_condition:
                    passing_where_clauses.append(f"NOT ({invalid_condition})")
                    where_clauses.append(f"({invalid_condition})")

            if self.filter:
                where_clauses.append(f"({self.filter})")
                passing_where_clauses.append(f"({self.filter})")

            where_sql = " AND ".join(where_clauses)
            passing_where_sql = " AND ".join(passing_where_clauses)

            sql = self.data_source_scan.data_source.sql_select_all(
                self.partition.table.table_name, self.samples_limit, where_sql
            )

            # Passing/failing queries are tied to existence of samples and sample query - this is ok for now as
            # whole Failed Rows Analysis is tied to existence of a failed rows sample file.
            if self.samples_limit > 0:
                sample_query = SampleQuery(self.data_source_scan, self, "failed_rows", sql)

                sample_query.passing_sql = self.data_source_scan.data_source.sql_select_all(
                    self.partition.table.table_name, filter=passing_where_sql
                )
                sample_query.failing_sql = self.data_source_scan.data_source.sql_select_all(
                    self.partition.table.table_name, filter=where_sql
                )

                return sample_query



from __future__ import annotations

from datetime import timedelta

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class FreshnessCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: str | None,
        location: Location,
        name: str | None,
        filter: str | None,
        column_name: str,
        variable_name: str | None,
        fail_freshness_threshold: timedelta,
        warn_freshness_threshold: timedelta,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.column_name: str = column_name
        self.variable_name: str = "NOW" if variable_name is None else variable_name
        self.fail_freshness_threshold: timedelta = fail_freshness_threshold
        self.warn_freshness_threshold: timedelta = warn_freshness_threshold
        self.filter = filter

    def get_column_name(self) -> str | None:
        return self.column_name



from dataclasses import dataclass
from typing import Optional


@dataclass(eq=True, frozen=True)
class NameFilter:
    data_source_name_filter: Optional[str]
    table_name_filter: Optional[str]
    partition_name_filter: Optional[str]
    column_name_filter: Optional[str]



from __future__ import annotations

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location
from soda.sodacl.threshold_cfg import ThresholdCfg


class UserDefinedFailedRowsExpressionCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: dict | None,
        location: Location,
        name: str,
        fail_condition_sql_expr: str | None,
        samples_limit: int | None = None,
        samples_columns: list | None = None,
        fail_threshold_cfg: ThresholdCfg | None = None,
        warn_threshold_cfg: ThresholdCfg | None = None,
    ):
        super().__init__(
            source_header,
            source_line,
            source_configurations,
            location,
            name,
            samples_limit=samples_limit,
            samples_columns=samples_columns,
        )
        self.fail_condition_sql_expr: str | None = fail_condition_sql_expr
        self.fail_threshold_cfg: ThresholdCfg | None = fail_threshold_cfg
        self.warn_threshold_cfg: ThresholdCfg | None = warn_threshold_cfg



#!/usr/bin/env python

from setuptools import find_namespace_packages, setup

package_name = "soda-core"
# Managed by tbump - do not change manually
package_version = "3.3.8"
description = "Soda Core"

requires = [
    "markupsafe>=2.0.1,<=2.1.2",
    "Jinja2>=2.11,<4.0",
    "click~=8.0",
    "ruamel.yaml>=0.17.0,<0.18.0",
    "requests~=2.30",
    "antlr4-python3-runtime~=4.11.1",
    "opentelemetry-api>=1.16.0,<1.23.0",  # 1.23.0 removes Python 3.7 support
    "opentelemetry-exporter-otlp-proto-http>=1.16.0,<1.23.0",
    "sqlparse~=0.4",
    "inflect~=7.0",
    "pydantic>=2.0.0,<3.0.0",
]

setup(
    name=package_name,
    version=package_version,
    author="Soda Data N.V.",
    author_email="info@soda.io",
    description="Soda Core library & CLI",
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["soda*"]),
    install_requires=requires,
    entry_points={"console_scripts": ["soda=soda.cli.cli:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.7",
)



from soda.execution.metric.query_metric import QueryMetric
from soda.execution.query.group_by_query import GroupByQuery


class GroupByMetric(QueryMetric):
    def __init__(self, data_source_scan: "DataSourceScan", partition: "Partition", check: "Check", query: str):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
            name="group by",
            check=check,
            identity_parts=[query],
        )
        self.query = query

    def ensure_query(self):
        check = next(iter(self.checks), None)
        location = check.check_cfg.location if check else None
        self.data_source_scan.queries.append(
            GroupByQuery(
                data_source_scan=self.data_source_scan,
                metric=self,
                location=location,
                partition=self.partition,
            )
        )



from typing import Dict, List

from soda.sodacl.data_source_scan_cfg import DataSourceScanCfg
from soda.sodacl.for_each_column_cfg import ForEachColumnCfg
from soda.sodacl.for_each_dataset_cfg import ForEachDatasetCfg


class SodaCLCfg:
    def __init__(self, scan):
        self.scan = scan
        self.data_source_scan_cfgs: Dict[str, DataSourceScanCfg] = {}
        self.for_each_dataset_cfgs: List[ForEachDatasetCfg] = []
        self.for_each_column_cfgs: List[ForEachColumnCfg] = []

    def get_or_create_data_source_scan_cfgs(self, data_source_name):
        data_source_scan_cfgs = self.data_source_scan_cfgs.get(data_source_name)
        if not data_source_scan_cfgs:
            data_source_scan_cfgs = DataSourceScanCfg(data_source_name)
            self.data_source_scan_cfgs[data_source_name] = data_source_scan_cfgs
        return data_source_scan_cfgs



from __future__ import annotations

import re

from soda.cloud.historic_descriptor import HistoricChangeOverTimeDescriptor
from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.group_evolution_metric import GroupEvolutionMetric
from soda.execution.metric.metric import Metric
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.group_evolution_check_cfg import (
    GroupEvolutionCheckCfg,
    GroupValidations,
)

KEY_GROUPS_MEASURED = "groups measured"
KEY_GROUPS_PREVIOUS = "groups previous"


# TODO - add support for cloud diagnostics and retrieving values from cloud
class GroupEvolutionCheck(Check):
    def __init__(
        self,
        check_cfg: CheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
        )
        self.cloud_check_type = "generic"
        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )

        check_cfg: UserDefinedFailedRowsCheckCfg = self.check_cfg

        metric = GroupEvolutionMetric(
            data_source_scan=self.data_source_scan,
            query=check_cfg.query,
            check=self,
            partition=partition,
        )
        metric = self.data_source_scan.resolve_metric(metric)
        self.metrics[KEY_GROUPS_MEASURED] = metric

        self.failures = 0
        self.warnings = 0
        self.failure_results = {"missing": [], "forbidden": [], "additions": [], "deletions": []}
        self.warning_results = {"missing": [], "forbidden": [], "additions": [], "deletions": []}

        group_evolution_check_cfg: GroupEvolutionCheckCfg = self.check_cfg
        if group_evolution_check_cfg.has_change_validations():
            historic_descriptor = HistoricChangeOverTimeDescriptor(
                metric_identity=metric.identity, change_over_time_cfg=ChangeOverTimeCfg()
            )
            self.historic_descriptors[KEY_GROUPS_PREVIOUS] = historic_descriptor

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        group_evolution_check_cfg: GroupEvolutionCheckCfg = self.check_cfg

        self.measured_groups = metrics.get(KEY_GROUPS_MEASURED).value

        previous_groups = (
            historic_values.get(KEY_GROUPS_PREVIOUS).get("measurements").get("results")[0].get("value")
            if historic_values
            and historic_values.get(KEY_GROUPS_PREVIOUS, {}).get("measurements", {}).get("results", {})
            else None
        )

        if previous_groups:
            self.group_comparator = GroupComparator(previous_groups, self.measured_groups)
        else:
            if group_evolution_check_cfg.has_change_validations():
                warning_message = "Skipping group checks since there are no historic groups available!"
                self.logs.warning(warning_message)
                self.add_outcome_reason(outcome_type="notEnoughHistory", message=warning_message, severity="warn")
                return

        if group_evolution_check_cfg.fail_validations:
            (self.failures, self.failure_results) = self.group_validations(group_evolution_check_cfg.fail_validations)
        if group_evolution_check_cfg.warn_validations:
            (self.warnings, self.warning_results) = self.group_validations(group_evolution_check_cfg.warn_validations)

        if self.failures > 0:
            self.outcome = CheckOutcome.FAIL
        elif self.warnings > 0:
            self.outcome = CheckOutcome.WARN
        else:
            self.outcome = CheckOutcome.PASS

    def group_validations(self, validations: GroupValidations) -> (int, dict):
        measured_groups = self.measured_groups
        required_groups = set()
        missing_groups = []
        present_groups = []

        if validations.required_group_names:
            required_groups.update(validations.required_group_names)

        if validations:
            for required_group_name in required_groups:
                if required_group_name not in measured_groups:
                    missing_groups.append(required_group_name)

        if validations.forbidden_group_names:
            for forbidden_group_name in validations.forbidden_group_names:
                regex = forbidden_group_name.replace("%", ".*").replace("*", ".*")
                forbidden_pattern = re.compile(regex)
                for group_name in measured_groups:
                    if forbidden_pattern.match(group_name):
                        present_groups.append(group_name)

        result = {"missing": missing_groups, "forbidden": present_groups, "additions": [], "deletions": set()}

        total_count = sum(len(result[key]) for key in result.keys())

        if validations.is_group_addition_forbidden and self.group_comparator:
            result["additions"] = self.group_comparator.group_additions
        if validations.is_group_deletion_forbidden and self.group_comparator:
            result["deletions"] = self.group_comparator.group_deletions

        return total_count, result

    def get_cloud_diagnostics_dict(self) -> dict:
        import inflect

        p = inflect.engine()
        failures = f"{self.failures} {p.plural('Failure', self.failures)}"
        warnings = f"{self.warnings} {p.plural('Warning', self.warnings)}"

        cloud_diagnostics = {
            "blocks": [],
            "preferredChart": "bars",
            "valueLabel": f"{failures}, {warnings}",
            "valueSeries": {
                "values": [
                    {"label": "fail", "value": self.failures, "outcome": "fail"},
                    {"label": "warn", "value": self.warnings, "outcome": "warn"},
                ]
            },
        }

        diagnostic_text = self.diagnoistics_text(self.warning_results, "warn") + self.diagnoistics_text(
            self.failure_results, "fail"
        )

        cloud_diagnostics["blocks"].append(
            {"type": "csv", "title": "Diagnostics", "text": f"Group, Reason\n{diagnostic_text}"}
        )

        return cloud_diagnostics

    def diagnoistics_text(self, results, result_type):
        return "\n".join(
            ([f":icon-{result_type}: {m}, Missing\n" for m in results["missing"]])
            + ([f":icon-{result_type}: {f}, Forbidden\n" for f in results["forbidden"]])
            + ([f":icon-{result_type}: {a}, Added\n" for a in results["additions"]])
            + ([f":icon-{result_type}: {d}, Deleted\n" for d in results["deletions"]])
        )


class GroupComparator:
    def __init__(self, previous_groups, measured_groups):
        self.group_additions = []
        self.group_deletions = []
        self.__compute_group_changes(previous_groups, measured_groups)

    def __compute_group_changes(self, previous_groups, measured_groups):
        for previous_group in previous_groups:
            if previous_group not in measured_groups:
                self.group_additions.append(previous_group)
        for group in measured_groups:
            if group not in previous_groups:
                self.group_additions.append(group)



from __future__ import annotations


class DataType:
    """
    Generic data types that data_sources can translate to specific types
    """

    TEXT = "text"
    INTEGER = "integer"
    DECIMAL = "decimal"
    DATE = "date"
    TIME = "time"
    TIMESTAMP = "timestamp"
    TIMESTAMP_TZ = "timestamptz"
    BOOLEAN = "boolean"

    @staticmethod
    def array(element_data_type: str) -> str:
        return f"array[{element_data_type}]"

    @staticmethod
    def struct(fields: dict[str, str]) -> str:
        field_data_types = ",".join(f"{field_name}:{field_type}" for field_name, field_type in fields.items())
        return f"struct[{field_data_types}]"



from __future__ import annotations

from datetime import timezone
from typing import Any

from soda.cloud.historic_descriptor import (
    HistoricCheckResultsDescriptor,
    HistoricMeasurementsDescriptor,
)
from soda.common.exceptions import SODA_SCIENTIFIC_MISSING_LOG_MESSAGE
from soda.execution.check.metric_check import MetricCheck
from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.metric.metric import Metric
from soda.execution.partition import Partition
from soda.sodacl.anomaly_detection_metric_check_cfg import (
    AnomalyDetectionMetricCheckCfg,
)

KEY_HISTORIC_MEASUREMENTS = "historic_measurements"
KEY_HISTORIC_CHECK_RESULTS = "historic_check_results"
HISTORIC_RESULTS_LIMIT = 1000


class AnomalyDetectionMetricCheck(MetricCheck):
    def __init__(
        self,
        check_cfg: AnomalyDetectionMetricCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None = None,
        column: Column | None = None,
    ):
        try:
            super().__init__(
                check_cfg=check_cfg,
                data_source_scan=data_source_scan,
                partition=partition,
                column=column,
            )
            self.check_cfg: AnomalyDetectionMetricCheckCfg
            self.skip_anomaly_check = False
            metric_name = self.check_cfg.metric_name
            metric = self.metrics[metric_name]
            self.historic_descriptors[KEY_HISTORIC_MEASUREMENTS] = HistoricMeasurementsDescriptor(
                metric_identity=metric.identity,
                limit=HISTORIC_RESULTS_LIMIT,
            )
            self.historic_descriptors[KEY_HISTORIC_CHECK_RESULTS] = HistoricCheckResultsDescriptor(
                check_identity=self.create_identity(), limit=HISTORIC_RESULTS_LIMIT
            )
            self.diagnostics = {}
            self.cloud_check_type = "anomalyDetection"
        except Exception as e:
            self.skip_anomaly_check = True
            data_source_scan.scan._logs.error(
                f"""An error occurred during the initialization of AnomalyMetricCheck. Please make sure"""
                f""" that the metric '{check_cfg.metric_name}' is supported. For more information see"""
                f""" the docs: https://docs.soda.io/soda-cl/anomaly-detection.html""",
                exception=e,
            )

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, dict[str, Any]]) -> None:
        if self.skip_anomaly_check:
            return

        if not isinstance(historic_values, dict):
            self.logs.error(
                "Getting historical measurements and check results from Soda Cloud resulted in a "
                f"{type(historic_values)} object which is not compatible with anomaly detection. "
                "Check previous log messages for more information."
            )
            return

        historic_check_results = historic_values.get(KEY_HISTORIC_CHECK_RESULTS, {}).get("check_results", {})
        historic_measurements = self.get_historic_measurements(metrics, historic_values)

        # TODO test for module installation and set check status to is_skipped if the module is not installed
        try:
            from soda.scientific.anomaly_detection_v2.anomaly_detector import (
                AnomalyDetector,
            )
        except ModuleNotFoundError as e:
            self.logs.error(f"{SODA_SCIENTIFIC_MISSING_LOG_MESSAGE}\n Original error: {e}")
            return

        anomaly_detector = AnomalyDetector(
            measurements=historic_measurements,
            check_results=historic_check_results,
            logs=self.logs,
            model_cfg=self.check_cfg.model_cfg,
            training_dataset_params=self.check_cfg.training_dataset_params,
            severity_level_params=self.check_cfg.severity_level_params,
        )
        level, diagnostics = anomaly_detector.evaluate()
        assert isinstance(diagnostics, dict), f"Anomaly diagnostics should be a dict. Got a {type(diagnostics)} instead"

        self.add_outcome_reason(
            outcome_type=diagnostics["anomalyErrorCode"],
            message=diagnostics["anomalyErrorMessage"],
            severity=diagnostics["anomalyErrorSeverity"],
        )
        self.diagnostics = diagnostics

        if diagnostics["anomalyErrorCode"] == "not_enough_measurements_custom":
            if self.diagnostics["value"] is None:
                self.diagnostics["value"] = self.get_metric_value()
            return
        self.outcome = CheckOutcome(level)

    def get_historic_measurements(
        self, metrics: dict[str, Metric], historic_values: dict[str, dict[str, Any]]
    ) -> dict[str, list[dict[str, Any]]]:
        metric_name = self.check_cfg.metric_name
        historic_measurements = historic_values.get(KEY_HISTORIC_MEASUREMENTS, {}).get("measurements", {})
        self.logs.debug(
            "Anomaly Detection: using historical measurements " f"for identity {metrics[metric_name].identity}"
        )
        if not historic_measurements:
            self.logs.warning(f"This is the first time that we derive {metrics[metric_name]} metric")
            historic_measurements["results"] = []

        # Append current results
        local_data_time = self.data_source_scan.scan._data_timestamp
        utc_data_time = local_data_time.astimezone(timezone.utc)
        utc_data_time_str = utc_data_time.strftime("%Y-%m-%dT%H:%M:%SZ")
        historic_measurements.get("results", []).append(
            {
                "id": "dummy_id",  # Placeholder number that will be overwritten
                "identity": metrics[metric_name].identity,
                "value": self.get_metric_value(),
                "dataTime": utc_data_time_str,
            }
        )
        return historic_measurements

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        return {**cloud_diagnostics, **self.diagnostics}

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        if self.historic_diff_values:
            log_diagnostics.update(self.diagnostics)
        return log_diagnostics

    def create_migrate_identities(self) -> dict[str, str] | None:
        """
        This method is used to migrate the identites from anomaly score to anomaly detection.
        It's a hack to obtain the same identity for the anomaly detection check as the anomaly score check.
        """
        if self.check_cfg.take_over_existing_anomaly_score_check is False:
            # Do not migrate identities if the flag is set to False
            return super().create_migrate_identities()
        original_source_line = self.check_cfg.source_line.strip()
        original_migrate_data_source_name = self.data_source_scan.data_source.migrate_data_source_name

        hacked_source_line = original_source_line.replace("anomaly detection", "anomaly score") + " < default"
        hacked_migrate_data_source_name = original_migrate_data_source_name
        if original_migrate_data_source_name is None:
            hacked_migrate_data_source_name = True

        self.check_cfg.source_line = hacked_source_line
        self.data_source_scan.data_source.migrate_data_source_name = hacked_migrate_data_source_name

        identities = super().create_migrate_identities()

        # Overwrite the original source line and migrate data source name to avoid confusion
        self.check_cfg.source_line = original_source_line
        self.data_source_scan.data_source.migrate_data_source_name = original_migrate_data_source_name
        return identities



from __future__ import annotations

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class GroupByCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: dict | None,
        location: Location,
        name: str,
        query: str,
        fields: dict,
        check_cfgs: list[CheckCfg] = [],
        group_limit: int = 1000,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.query = query
        self.fields = fields
        self.check_cfgs = check_cfgs
        self.group_limit = group_limit
